// Copyright (C) 1996 Yale Institute for Biospheric Studies and Ginger Booth.
// This library is distributed without any warranty; without even the
// implied warranty of merchantability or fitness for a particular purpose.
// See file LICENSE for details and terms of copying.

// vfb, original, 960620:  copy from HeatbugBatchSwarm and fit to Gecko.

#import "Base.h"
#import "GeckoModelSwarm.h"
#import "GeckoDisplaySwarm.h"
#import "GeckoControlSwarm.h"
#import <time.h>
#import <simtools.h>
#import <activity.h>
#import <tkobjc.h>
#import "QuadDisplay.h"
#import "OtherDisplay.h"
#import "Interpret.h"
#import "Evolve.h"

@implementation ControlParams

-makeProbeMap 
{
    ProbeMap * probeMap;
    
    probeMap = [CompleteProbeMap createBegin: [self getZone]];
    [probeMap setProbedClass: [self class]];
    probeMap = [probeMap createEnd];
    return self;
}

-setStatCycle: (int) i { statcycle = i; return self; }
-setTimeStep: (int) i { timestep = i; return self; }
-setDisplayCycle: (int) i { displaycycle = i; return self; }
-setGeckoRuns: (int) i { geckoRuns = i; return self; }
-setESRun: (int) i { ESRun = i; return self; }
-setFileStem: (char*) s { fileStem = s; return self; }

-addParams
{
    [super init];
    [self addParam: "StatCycle" Vary: 0 IValue: statcycle 
          Message: M(setStatCycle:) Owner: self];
    [self addParam: "ButtonTimeStep" Vary: 0 IValue: timestep
          Message: M(setTimeStep:) Owner: self];
    [self addParam: "Display" Vary: 0 IValue: displaycycle
          Message: M(setDisplayCycle:) Owner: self];
    [self addParam: "Runs" Vary: 0 IValue: geckoRuns
          Message: M(setGeckoRuns:) Owner: self];
    [self addParam: "ESRun" Vary: 0 IValue: ESRun
          Message: M(setESRun:) Owner: self];
    [self addParam: "OutputStem" Vary: 0 SValue: fileStem
          Message: M(setFileStem:) Owner: self];
        
    return self;
}

+create
{
    ControlParams* obj;
    
    obj = [[ControlParams createBegin: globalZone] createEnd];
    [obj setName: "RunControl"];
    obj->statcycle    = 10;
    obj->timestep     = 1;
    obj->displaycycle = 100;
    obj->geckoRuns    = 1;
    obj->ESRun        = 0;
    obj->fileStem     = "gecko";

    [obj addParams];
    [obj makeProbeMap];
    
    return obj;
}

@end

@implementation GeckoControlSwarm

-setFileStem: (char*) stem
    // stem.log, stem.arc, stem.popGraph, etc.
{ settings->fileStem = stem; return self; }
-(char*) getFileStem { return settings->fileStem; }
-setDisplayFlag: (int) df {displayFlag = df; return self;}
-(BOOL) getDisplayFlag {return displayFlag;}
-setStatCycle: (int) sc {settings->statcycle = sc; return self;}
-setGeckoRuns: (int) gr {settings->geckoRuns = gr; return self;}
-(int) getGeckoRuns {return settings->geckoRuns;}
-(GeckoModelSwarm*) getModelSwarm {return geckoModelSwarm;}
-(FILE*) getLogFile { return logFile; }
-(FILE*) getArchiveFile { return archiveFile; }
-setESRun {settings->ESRun = 1; return self; }
-(BOOL) getESRun { return settings->ESRun; }
-(int) getStatCycle { return settings->statcycle; }
-setProtocol: (char*) s { protocolFileName = s; return self; }
-(FILE*) getProtocol {return protocolFile;}

-init
    // section out the re-initialization
{
    if (popGraph)
        [popGraph clear];
    if (biomassGraph)
        [biomassGraph clear];
    
    return self; 
}


+createBegin: (id) aZone 
    // set defaults
{
    GeckoControlSwarm * obj;

    obj = [super createBegin: aZone];
    [obj init];
    
    obj->settings = [ControlParams create];
    obj->displayFlag  = 1; // unless proven otherwise....
    
    obj->protocolFileName = "";
    obj->protocolFile   = 0;
    
    obj->popGraph    	= nil;
    obj->biomassGraph   = nil;

    return obj;
}

-buildStatGraphs1
    // before model build
{
    char namebuf[50];
    
    popGraph = [TraceGraph create];

    sprintf(namebuf, "%s%s", settings->fileStem, ".popGraph");
    [popGraph logFile: namebuf];
    [popGraph setWinTitle: "Population"];
    biomassGraph = [TraceGraph create];
    
    sprintf(namebuf, "%s%s", settings->fileStem, ".biomassGraph");
    [biomassGraph logFile: namebuf];
    [biomassGraph setWinTitle: "Biomass"];

    return self;    
}

-buildStatGraphs2
    // after model build
{
    // the traces need to be created per-species
    popGraph = [popGraph createEnd];
    if (displayFlag) {
        [popGraph->graph setWindowGeometry: "500x310+36+0"];
        [popGraph display];
    }
    biomassGraph = [biomassGraph createEnd];
    if (displayFlag) {
        [biomassGraph->graph setWindowGeometry: "500x310+36+353"];
        [biomassGraph display];
    }

    return self;    
}


-buildObjects 
{
    id pdSelf, pdModel;				// popup handlers
    id modelZone;				// zone for model.
    id displayZone;				// zone for graphics.
    char namebuf[50];
    char* whereami;

    [super buildObjects];
    [self init];

    // create a zone for the model, create the model there.
    modelZone = [Zone create: [self getZone]];
    geckoModelSwarm = [GeckoModelSwarm create: modelZone];

    // read control file.  
    if (strcmp(protocolFileName, "")) {
        protocolFile = fopen(protocolFileName, "r");
        if (!protocolFile) {
            geckoError(InvalidArgument, "Unable to open protocol file: %s.\n", 
                        protocolFileName);
        } else {
            whereami = [settings loadToNextSection: protocolFile];
            if (!strcmp(whereami, "RunControl")) {
                geckoError(InvalidArgument, "Expected 'RunControl' in protocol.\n");
            } else {
                [settings loadToNextSection: protocolFile];
            }
        }
    }

    if (displayFlag) {
        displayZone = [Zone create: [self getZone]];
        geckoDisplaySwarm = [GeckoDisplaySwarm createBegin: displayZone];
        [geckoDisplaySwarm setModel: geckoModelSwarm];
        geckoDisplaySwarm = [geckoDisplaySwarm createEnd];

        // popup some menus to allow user to intervene
        if (!protocolFile) {
            pdSelf = [probeDisplayManager createProbeDisplayFor: settings];
            pdModel = [probeDisplayManager createProbeDisplayFor: 
                    geckoModelSwarm->settings];
        }
     
        controlPanel = [ControlPanel create: [self getZone]];
        [[controlPanel getPanel] setPositionX:-1 Y: 706];
        [controlPanel waitForControlEvent];
        if ([controlPanel getState] == ControlStateQuit)
    
            return self;
        
        // now get rid of the popups, since editing them is useless.
        if (!protocolFile) {
            [pdSelf drop];
            [pdModel drop];
        }
        
        // displaycycle belongs downstream, but can't abide 3rd parm popup
        [geckoDisplaySwarm setDisplaycycle: settings->displaycycle];
    }

    // let's open the log files ASAP
    sprintf(namebuf, "%s%s", settings->fileStem, ".log");
    logFile = fopen(namebuf, "w");
    sprintf(namebuf, "%s%s", settings->fileStem, ".out");
    archiveFile = fopen(namebuf, "w");
    
    // Let the model swarm build its objects.
    [self buildStatGraphs1];
    [geckoModelSwarm buildObjects];
    [self buildStatGraphs2];

    // And echo the input to output, in all cases.  Not yet....
    if (!settings->ESRun)
        [self echoInput];

    // and the display swarm - needed model first.
    if (displayFlag) {
        [geckoDisplaySwarm buildObjects];
    }    
    
    return self;
}  


// Create the actions necessary for the simulation. This is where
// the schedule is built (but not run!)
-buildActions {
    [super buildActions];
    
    // First, let our model swarm build its own schedule.
    [geckoModelSwarm buildActions];

    reportActions = [ActionGroup create: [self getZone]];
    [reportActions createActionTo: geckoModelSwarm message: M(doStats)];
    // note these graphs step at statcycle, not displaycycle
    [reportActions createActionTo: popGraph message: M(step)];
    [reportActions createActionTo: biomassGraph message: M(step)];
    if (displayFlag) {
        [reportActions createActionTo: controlPanel   message: M(doTkEvents)];
    }
    [reportActions createActionTo: geckoModelSwarm message: M(extinctCheck)];
    
    if (displayFlag)
        [geckoDisplaySwarm buildActions];
    
    // the reportSchedule controls how often we write data out.
    reportSchedule = [Schedule createBegin: [self getZone]];
    [reportSchedule setRepeatInterval: settings->statcycle];
    reportSchedule = [reportSchedule createEnd];
    [reportSchedule at: 0 createAction: reportActions];
    
    // we also add in a "stopSchedule", another schedule with an absolute
    // time event - stop the system at stopday. 
    stopSchedule = [Schedule create: [self getZone]];
    [stopSchedule at: [geckoModelSwarm getStopDay]
         createActionTo: self message: M(stopRunning)];
    
    return self;
} 

// activateIn: - get the Swarm ready to run.
-activateIn: (id) swarmContext {
    // First, activate ourselves (just pass along the context).
    [super activateIn: swarmContext];

    // And, we need to activate the subswarms.
    [geckoModelSwarm activateIn: self];
    if (displayFlag)
        [geckoDisplaySwarm activateIn: self];
        
    // Now activate our schedules in ourselves.
    [reportSchedule activateIn: self];
    [stopSchedule activateIn: self];
    
    // Activate returns the swarm activity - the thing that's ready to run.
    return [self getSwarmActivity];
}
 

-go 
    // may need to run the control panel, if display on.
{
    if (!(displayFlag)) {
        [swarmActivity run];
        return [swarmActivity getStatus];
    }
    // now the hard part - run the control panel
    
    // infinite loop: this loop only exits on a return statement, either
    //   controlStateQuit: the user hit "quit" before the activities were done
    //   Completed: the activity library indicated it was done running us
    while (1) {
        id controlState, swarmStatus;
        // These variables indicate the running state: both the activity
        // library's notion of where we are, as well as the control panel.
        swarmStatus = [swarmActivity getStatus];
        controlState = [controlPanel getState];

        // First check if we can exit because our Swarm thinks we're done.
        if (swarmStatus == Completed) {
            // We're ready to return a value - just check if the user hit "quit".
            if (controlState == ControlStateQuit) {
	        // Ok, the user has hit the quit button while we were running.
                [self setGeckoRuns: 0];
	        return ControlStateQuit;
            } else {
	        // Looks like the schedule ran to completion, or at least exited
	        // for some reason other than the user hitting "Quit".
	        return Completed;
            }
        } else {
            // The activity library has not finished executing - that means we
            // need to arrange for some other sort of action.
            if (controlState == ControlStateRunning) {
	        // Control panel says we're ready to run, so run!
	        
	        [swarmActivity run];
	    
            } else if (controlState == ControlStateStopped) {
	        // Control panel says we're stopped, so let's wait for a go.
	        
	        [controlPanel waitForControlEvent];
	    
            } else if (controlState == ControlStateQuit) {
	        // Control panel says to quit, even though activity hasn't. Quit.

	        return ControlStateQuit;

            } else if (controlState == ControlStateStepping) {
	        [swarmActivity 
	            stepUntil: [swarmActivity getCurrentTime] 
	                       + settings->timestep];
	        [controlPanel setStateStopped];

	    // this one appears to control "Time Step" button
            } else if (controlState == ControlStateNextTime) {
	        [swarmActivity 
	            stepUntil: [swarmActivity getCurrentTime] 
	                       + settings->timestep];
	        [controlPanel setStateStopped];
            } else {
	        [controlPanel setStateStopped];
            }
        }	
  }	

  // That while loop should never exit, so just put a check here.
  geckoError(SourceMessage, "The control Swarm go loop exited improperly.\n");
  return nil;
}

-resetClear 
{
    [popGraph clear];
    [biomassGraph clear];
    
    [geckoModelSwarm resetClear];
    
    // must call after model reset to get latest worldRoot.
    if (displayFlag)
        [geckoDisplaySwarm resetClear];
    
    return self;
}

// And the termination method. When this fires we terminate everything
// that's running and close our output files, if no more runs to do.
// If more runs to do, close this one and reset.
-stopRunning 
{
    [getTopLevelActivity() terminate];

    if (settings->ESRun) {
        if ([scenariosToDo nextScenario])
            [self setGeckoRuns: 1];
        else
            [self setGeckoRuns: 0];
    } else
        [self setGeckoRuns: [self getGeckoRuns] - 1];
    
    // all done?
    if ([self getGeckoRuns] == 0) {
        fclose(logFile);
        fclose(archiveFile);
    // restart
    } else {
        flushBoth();
        flushBoth();
        [self resetClear];
    }
    return self;
}


-echoInput 
{
    printBoth("# %s\n", GeckoVersion);
    printBoth("# Run begins %s\n", timeString());
    printBoth("# Input:\t\t%s\n\n", protocolFileName);
    printBoth("@begin RunControl\n");
    [settings dump];
    printBoth("@end RunControl\n\n");

    [geckoModelSwarm dump];
    
    return self;
}

-printv
{
    [geckoModelSwarm printv];
    return self;
}

-println
{
    [geckoModelSwarm println];
    return self;
}

-printCensus
{
    [geckoModelSwarm printCensus];
    return self;
}

@end
