// Copyright (C) 1996 Yale Institute for Biospheric Studies and Ginger Booth.
// This library is distributed without any warranty; without even the
// implied warranty of merchantability or fitness for a particular purpose.
// See file LICENSE for details and terms of copying.

// vfb, original, 960618:  copy from HeatbugModelSwarm and fit to Gecko.

#import "GeckoModelSwarm.h"
#import "GeckoControlSwarm.h"
#import "LandRoot.h"
#import "ResourceQuad.h"
#import "FeedQuad.h"
#import "Evolve.h"
#import "CritterZoo.h"
#import "GrazerZoo.h"
#import "CarnieZoo.h"
#import "Fission2D.h"
#import "random.h"
#import <simtools.h>

double curRunFitness = 0.0;

@implementation ModelParams

-makeProbeMap 
{
    ProbeMap * probeMap;
    
    probeMap = [CompleteProbeMap createBegin: [self getZone]];
    [probeMap setProbedClass: [self class]];
    probeMap = [probeMap createEnd];
    return self;
}

-setStopDay: (int) i { stopday = i; return self; }
-setReQuad: (int) i { reQuadRound = i; return self; }
-setSeason: (int) i { season = i; return self; }
// unfortunately, the random seed is a long int, and we have no specific
// support for that.  Hence, a string.
-setRandomSeed: (char*) s 
{ 
    char* sbuf;
    long longint;
    
    sscanf(s, "%ld", &longint); 
    if ((longint == 0) || (longint == -1)) {
        longint = NO_SEED;
        longint = InitRand(longint);
        sbuf = [globalZone alloc: 50];
        sprintf(sbuf, "%ld", longint); 
        [self setParam: "RandomSeed" SValue: sbuf];
    }
    longint = InitRand(longint);
    return self; 
}

-(long) getRandomSeed {
    long longint;
    Param* randparam = [self getParam: "RandomSeed"];
    sscanf([randparam getSValue], "%ld", &longint); 
    return longint;
}

-addParams
{
    char* sbuf;
    
    [super init];
    sbuf = [globalZone alloc: 50];
    sprintf(sbuf, "%s", "-1");
    [self addParam: "RandomSeed" Vary: 0 SValue: sbuf
          Message: M(setRandomSeed:) Owner: self];
    [self addParam: "StopDay" Vary: 0 IValue: stopday
          Message: M(setStopDay:) Owner: self];
    [self addParam: "ReQuad" Vary: 0 IValue: reQuadRound
          Message: M(setReQuad:) Owner: self];
    [self addParam: "Season" Vary: 0 IValue: season
          Message: M(setSeason:) Owner: self];
    return self;
}

+create
{
    ModelParams* obj;
    
    obj = [[ModelParams createBegin: globalZone] createEnd];
    [obj setName: "World"];
    obj->stopday      = 3000;
    obj->reQuadRound  = 10;
    obj->season       = 0;
    [obj addParams];
    [obj makeProbeMap];
    
    return obj;
}

@end

@implementation GeckoModelSwarm

Mcr*	birthplace1;
Mcr*    birthplace2;
Mcr*    birthplace3; 

-(Quad*) getWorldRoot { return worldRoot; }

-(LandRoot*) getLandRoot { return landRoot; }
-(Quad*) getMobileRoot { return mobileRoot; }
-(Quad*) getPlantRoot { return plantRoot; }
-(int) getStopDay { return settings->stopday; }
-(int) getReQuadRound { return settings->reQuadRound; }
-(int) getCarniecycle { return carniecycle; }
-(id) getSpecies { return species; }
-(id) getSites { return sites; }
-setStopDay: (int) sd { settings->stopday = sd; return self; }

// createBegin: here we set up the default simulation parameters.
+createBegin: (id) aZone {
    GeckoModelSwarm * obj;

    // First, call our superclass createBegin - the return value is the
    // allocated GeckoModelSwarm object.
    obj = [super createBegin: aZone];

    // Now fill in various simulation parameters with default values.
    obj->plantcycle 	 = 1;
    obj->carniecycle 	 = 1;
    obj->settings        = [ModelParams create];

    obj->species         = [List create: aZone];
    obj->sites           = [List create: aZone];
    
    // world data structures
    obj->worldRoot 	= nil;
    obj->landRoot 	= nil;
    obj->mobileRoot 	= nil;
    obj->plantRoot 	= nil;
  
    return obj;
}

-initGlobals 
{
    // initialize Resource manager
    setNumRes(6);

    // some kludgy scaffolding 
    spidersReproducedFlag = 0;
    
    return self;
}

-(long) getRandomSeed {
    return [settings getRandomSeed];
}

-birthplaceReset
{
    Site* siteList;
    ResourceQuad* site1;
    GCoord halfx;

    siteList = [sites first];
    site1 = [siteList getGrandma];
    halfx = site1->mcr->ux / 2.0;

    // lower left 
    birthplace1->lx = 0;
    birthplace1->ly = 0;
    birthplace1->ux = halfx;
    birthplace1->uy = halfx * 4.0 / 3.0;
    
    // lower right
    birthplace2->lx = halfx;
    birthplace2->ly = 0;
    birthplace2->ux = 2.0 * halfx;
    birthplace2->uy = birthplace1->uy;
    
    // top
    birthplace3->lx = 0;
    birthplace3->ly = birthplace1->uy;
    birthplace3->ux = 2.0 * halfx;
    birthplace3->uy = 2.0 * halfx;

    return self;
}

-loadLand: (FILE*) pf 
{
    char* wordp;
    int wordlen, i;
    Site* onesite;
    char* string;
    id grandma;
    char copy5[5];

    // first site required.
    wordp = [settings loadWhere];
    for (i=0; i < 5; i++) copy5[i] = wordp[i]; copy5[4] = '\0';
    if (strcmp(copy5, "Site")) {
        geckoError(InvalidArgument,
            "Expected 'Site' in protocol: %s\n", [settings loadWhere]);
        return self;
    }

    // while more sites....
    
    while (!strcmp(copy5, "Site")) {
        wordp = [settings loadNextWord: pf];
        if (!(wordp = [settings loadNextWord: pf])) {
            geckoError(InvalidArgument,  "Unexpected end of protocol.\n");
            return self;
        }
            
        // get out new site's name
        wordlen = 0;
        while (!isspace(wordp[wordlen])) wordlen++;
        wordlen++;
        string = [[self getZone] alloc: wordlen + 1];
        for (i=0; i <= wordlen; i++) 
            string[i] = wordp[i];
        onesite = [Site create];
        [onesite setName: string];
        
        // get out new site type
        if (!(wordp = [settings loadNextWord: pf])) {
            geckoError(InvalidArgument,  "Unexpected end of protocol.\n");
            return self;
        }
        if (strcmp(wordp, "Type")) {
            geckoError(InvalidCombination,  
                "Site requires 'Type' in protocol, found: '%s'\n", wordp); 
            return self;
        }
        if (!(wordp = [settings loadNextWord: pf])) {
            geckoError(InvalidArgument,  "Unexpected end of protocol.\n");
            return self;
        }
        if (!objc_lookup_class(wordp)) {
            geckoError(InvalidArgument, 
                "Invalid Site Type = '%s' in protocol.\n", wordp );
             return self;
        }

        grandma = [objc_lookup_class(wordp) create];
        onesite = [onesite make: string WithGrandma: grandma];
        [onesite loadToNextSection: pf];
        [sites addLast: onesite];
        wordp = [settings loadWhere];
        for (i=0; i < 5; i++) copy5[i] = wordp[i]; copy5[4] = '\0';
    }
    return self;
}

-loadSpecies: (FILE*) pf 
{
    char* wordp;
    int wordlen, i;
    Species* onespecies;
    char* string;
    id grandma;
    char copy8[8];

    // first species required.
    wordp = [settings loadWhere];
    for (i=0; i < 8; i++) copy8[i] = wordp[i]; copy8[7] = '\0';
    if (strcmp(copy8, "Species")) {
        geckoWarning("Expected 'Species' in protocol: %s\n", 
            [settings loadWhere]);
    }

    // while more species....
    
    while (!strcmp(copy8, "Species")) {
        wordp = [settings loadNextWord: pf];
        if (!(wordp = [settings loadNextWord: pf])) {
            geckoError(InvalidArgument,  "Unexpected end of protocol.\n");
            return self;
        }
            
        // get out new species' name
        wordlen = 0;
        while (!isspace(wordp[wordlen])) wordlen++;
        wordlen++;
        string = [[self getZone] alloc: wordlen + 1];
        for (i=0; i <= wordlen; i++) 
            string[i] = wordp[i];
        
        // get out new species' type
        if (!(wordp = [settings loadNextWord: pf])) {
            geckoError(InvalidArgument,  "Unexpected end of protocol.\n");
            return self;
        }
        if (strcmp(wordp, "Type")) {
            geckoError(InvalidCombination,  
                "Species requires 'Type' in protocol, found: '%s'\n", wordp); 
            return self;
        }
        if (!(wordp = [settings loadNextWord: pf])) {
            geckoError(InvalidArgument,  "Unexpected end of protocol.\n");
            return self;
        }
        if (!objc_lookup_class(wordp)) {
            geckoError(InvalidArgument, 
                "Invalid Species Type = '%s' in protocol.\n", wordp );
             return self;
        }

        onespecies = [Species create];
        [onespecies setName: string];
        grandma = [objc_lookup_class(wordp) create];
        onespecies = [onespecies make: string WithGrandma: grandma
                       Color: 1
                       Birthday: 1 InitNum: 1
                       Mobile: 1 BioScale: 1.0];
        [onespecies loadToNextSection: pf];
        [onespecies createEnd];
        [species addLast: onespecies];
        wordp = [settings loadWhere];
        for (i=0; i < 8; i++) copy8[i] = wordp[i]; copy8[7] = '\0';
    }
    return self;
}

-loadCompositeSpecies: (FILE*) pf 
{
    char* wordp;
    int wordlen, i;
    CompositeSpecies* composite;
    Species* onespecies;
    char* string;
    char copy17[17];

    // no composite species required.
    wordp = [settings loadWhere];
    for (i=0; i < 17; i++) copy17[i] = wordp[i]; copy17[16] = '\0';
    if (strcmp(copy17, "CompositeSpecies")) {
        return self;
    }

    // while more composite species....
    
    while (!strcmp(copy17, "CompositeSpecies")) {
        wordp = [settings loadNextWord: pf];
        if (!(wordp = [settings loadNextWord: pf])) {
            geckoError(InvalidArgument,  "Unexpected end of protocol.\n");
            return self;
        }
            
        // get out composite name
        wordlen = 0;
        while (!isspace(wordp[wordlen])) wordlen++;
        wordlen++;
        string = [[self getZone] alloc: wordlen + 1];
        for (i=0; i <= wordlen; i++) 
            string[i] = wordp[i];
        composite = [CompositeSpecies create];
        [composite setName: string];

        wordp = [settings loadNextWord: pf];
        // parse the member species
        do {
            for (i=0; i < 6; i++) copy17[i] = wordp[i]; copy17[6] = '\0';
            if (!strcmp(copy17, "Member")) {
                if (!(wordp = [settings loadNextWord: pf])) {
                    geckoError(InvalidArgument,  "Unexpected end of protocol.\n");
                    return self;
                }
                
                // settings is irrelevant here, just something Parameterized

                if ((onespecies = (Species*) [settings getParameterized: wordp])) {
                    [composite addSpecies: onespecies];
                } else {
                    geckoError(InvalidArgument, 
                        "Invalid composite member species = '%s'.\n", wordp );
                    return self;
                }
                wordp = [settings loadNextWord: pf];
            }
        } while (!strcmp(copy17, "Member"));

        [composite reloadln: pf];
        [composite createEnd];
        [species addLast: composite];

        // another composite?
        wordp = [settings loadWhere];
        for (i=0; i < 17; i++) copy17[i] = wordp[i]; copy17[16] = '\0';
    }
    return self;
}

-makeSpecies
    // want this to move to Protocol, but get it to work first....
{
    Critter* pGrandma;
    Grazer*  gGrandma;
    Carnie*  spiderGrandma;
    Species *species1, *species2, *species3;
    
    // declare my species  
    pGrandma = [Critter create];
    species1 = [Species create];
    species1 = [species1 make: "Plant" WithGrandma: pGrandma
                       Color: 1
                       Birthday: 0 InitNum: 200
                       Mobile: 0 BioScale: 5.4];
    [species1 setParam: "BreedSize" DValue: 5.0];
    [species1 observePop];
    [species1 observeBiomass];
    [species addLast: species1];

    gGrandma = [Grazer create];
    species2 = [Species create];
    species2 = [species2 make: "Herbivore" WithGrandma: gGrandma
                       Color: GrazerColor
                       Birthday: 200  InitNum: 15
                       Mobile: 1 BioScale: 1.8];
    [species2 observePop];
    [species2 observeBiomass];
    [species addLast: species2];

    spiderGrandma = [Carnie create];
    species3 = [Species create];
    species3 = [species3 make: "Carnivore" WithGrandma: spiderGrandma
                       Color: CarnieColor
                       Birthday: 291  InitNum: 3
                       Mobile: 1 BioScale: 1.0];
    [species3 observePop];
    [species3 observeBiomass];
    [species addLast: species3];

    return self;
}

-makeRoots
{
    worldRoot = [Quad create];
    landRoot = [LandRoot create];
    landRoot->type = landscape;
    landRoot->parent = worldRoot;
    landRoot->sibling = worldRoot->child;
    worldRoot->child = landRoot;

    plantRoot = [Quad create];
    plantRoot->type = sessile;
    plantRoot->sibling = landRoot->sibling;
    landRoot->sibling  = plantRoot;
    plantRoot->parent  = worldRoot;

    mobileRoot = [Quad create];
    mobileRoot->type = mobile;
    plantRoot->sibling = mobileRoot;
    mobileRoot->sibling = nil;
    mobileRoot->parent  = worldRoot;
    return self;
}

-makeLand
{
    FeedQuad* 	sitedef;
    Site* 	sitemaster;

    // declare my sites    
    sitedef = [FeedQuad create];
    sitemaster = [Site create];
    sitemaster = [sitemaster make: "OnlySite" WithGrandma: sitedef ];
    [sitemaster setParam: "Limit" DValue: 1.0];
    [sitemaster setParam: "McrUX" DValue: 250.0];
    [sitemaster setParam: "McrUY" DValue: 250.0];
    [sitemaster varyParam: "McrUX"];
    [sitemaster offsetParam: "McrUY" Off: [sitemaster getParam: "McrUX"]]; 
    [sites addLast: sitemaster];
    
    return self;
}

-hangSites
{
    id 		sitegrandma;
    ResourceQuad* realsite;
    Site* 	site;
    id     	index;

    index = [sites begin: [self getZone]];

    while ( (site = [index next] ) ) {
        sitegrandma = [site getGrandma];
        realsite = [sitegrandma copymom];
        realsite->parent = landRoot;
        realsite->sibling = landRoot->child;
        landRoot->child = realsite;
        [realsite growMcr];
    }
    return self;
}

-hangCritters: (Critter*) grandma Num: (int) num 
         Root: (Quad*) root Where: (Mcr*) where
    // scatters num copies of grandma across the field, hanging from root.
    // Default all new critters to size...kidsize.  Only strews critters
    // on site 1 for now if no where specified.
{
    int i;
    Critter* one;
    Site* siteList;
    ResourceQuad* site1;
    GCoord x, y;
    GCoord x0, y0;
    GCoord x1, y1;

    [root squashQuad];
    
    if (!where) {
        siteList = [sites first];
        site1 = [siteList getGrandma];
        x0 = site1->mcr->lx;
        y0 = site1->mcr->ly;
        x1 = site1->mcr->ux;
        y1 = site1->mcr->uy;
    } else {
        x0 = where->lx;
        y0 = where->ly;
        x1 = where->ux;
        y1 = where->uy;
    }

    // Create critters themselves and hang on the quad tree.
    for (i = 0; i < num; i++) {
        one = [grandma copymom];
        x = x0 + EvenRand() * (x1 - x0);
        y = y0 + EvenRand() * (y1 - y0);
        [one setX: x Y: y Radius: grandma->kidsize];
        [root addNode: one->myquad];
    }
    [root subQuad];
    [root growMcr];

    return self;
}

-birthdayCheck
{
    [species forEach: M(birthdayCheck)];
    return self;
}
-deathdayCheck
{
    [species forEach: M(deathdayCheck)];
    return self;
}
-onSeason
{
    if ( (settings->season > 0) && 
         ( (getCurrentTime() % settings->season) == 0 ) ) {
        [sites forEach: M(onSeason)];
        [species forEach: M(onSeason)];
    }
    return self;
}

-resetClear
    // Since this hangs sites, should set up next run before calling.
{
    [worldRoot delTree];
    [species forEach: M(resetClear)];
    [self makeRoots];
    [self hangSites];
    [sites forEach: M(zeroBiomass)];
    return self;
}

-printv
{
    [species forEach: M(printv)];
    [sites forEach: M(printv)];
    return self;
}

-println
{
    [species forEach: M(println)];
    [sites forEach: M(println)];
    return self;
}
-printCensus
{
    [species forEach: M(printCensus)];
    printLog("\n");
    return self;
}

-dump
    // executable
{
    printBoth("@begin World\n");
    [settings dump];
    printBoth("@end World\n\n");
    [sites forEach: M(dump)];
    [species forEach: M(dump)];
    [scenarios dump];
    return self;
}


// Now it's time to build the model objects. We use various parameters
// inside ourselves to choose how to create things.
-buildObjects {
    FILE* pf;
    char copy18[18];
    char* wordp;
    int i;
    
    // allow our parent class to build anything.
    [super buildObjects];

    // divide and conquer the building processes...
    [self initGlobals];

    if ((pf = [geckoControl getProtocol])) {
        wordp = [settings loadWhere];
        for (i=0; i < 6; i++) copy18[i] = wordp[i]; copy18[5] = '\0';
        if (strcmp(copy18, "World")) {
            geckoError(InvalidArgument, 
                "Expected 'World' in protocol, found '%s'.\n", 
                [settings loadWhere]);
        } else {
            [settings loadToNextSection: pf];
        }
        [self makeRoots];
        for (i=0; i < 5; i++) copy18[i] = wordp[i]; copy18[4] = '\0';
        if (!strcmp(copy18, "Site")) {
            [self loadLand: pf];
            [self loadSpecies: pf];
        } else {
            [self loadSpecies: pf];
            [self loadLand: pf];
        }
        [self loadCompositeSpecies: pf];
    } else {
        [self makeRoots];
        [self makeLand];
        [self makeSpecies];
    }
    [self hangSites];

    if (pf) {
        wordp = [settings loadWhere];
        for (i=0; i < 18; i++) copy18[i] = wordp[i]; copy18[17] = '\0';
    }
    if ( [geckoControl getESRun] || 
         ( !strcmp(copy18, "EvolutionStrategy") ) ) {
        [scenariosToDo drop];             
        scenariosToDo = [ESTable create];
        scenarios = scenariosToDo;
        
        if ( (pf) &&
             ( !strcmp(copy18, "EvolutionStrategy") ) ) {
            [geckoControl setESRun];
            [scenarios loadToNextSection: pf];
        } else {
            printLog("# *** Using default values for Evolution Strategy.\n");
            [scenariosToDo setNumTrials: 3];
            [scenariosToDo setNumOffspring: 10];
            [scenariosToDo setTweakSigma: 2];
        }
        [scenariosToDo setFitFunctionOwner: self];
        [scenariosToDo setFitFunction: M(runFitness:)];
        scenariosToDo = [scenariosToDo createEnd];
        // [scenariosToDo setInitFitness:  1.1]; // true
        [scenariosToDo start];
    }
    
    return self;
}

// Here is where the model schedule is built, the data structures
// that define the simulation of time in the model. The core is an
// actionGroup that has a list of actions. That's then put in a Schedule.
-buildActions {

    [super buildActions];
    
    // Create the list of simulation actions. We put these in an action
    // group, because we want these actions to be executed in a specific
    // order, but these steps should take no (simulated) time. The
    // M(foo) means "The message called <foo>". You can send a message
    // To a particular object, or ForEach object in a collection.
    
    modelActions = [ActionGroup create: [self getZone]];

    // note the order of these 3 is crucial to seasons
    [modelActions createActionTo: self      message: M(deathdayCheck)];
    [modelActions createActionTo: self      message: M(onSeason)];
    [modelActions createActionTo: self      message: M(birthdayCheck)];
    
    [modelActions createActionTo: landRoot  
                  message: M(stepResourcesOn:) : (id)plantcycle];
    [modelActions createActionTo: landRoot  
                  message: M(feedTree:On:) : plantRoot : (id)plantcycle];
    [modelActions createActionTo: plantRoot    
                  message: M(stepOn:) : (id) plantcycle];
    [modelActions createActionTo: mobileRoot 
                  message: M(stepOn:) : (id) 1];

    // This is a simple schedule, with only one action that is just
    // repeated every time. 
    
    modelSchedule = [Schedule createBegin: [self getZone]];
    [modelSchedule setRepeatInterval: 1];
    modelSchedule = [modelSchedule createEnd];
    [modelSchedule at: 0 createAction: modelActions];
    
    return self;
}

// Now set up the model's activation. swarmContext indicates where
// we're being started in - typically, this model is run as a subswarm
// of an observer swarm.
-activateIn: (id) swarmContext {
  // First, activate ourselves via the superclass activateIn: method.
  // Just pass along the context: the activity library does the right thing.
  [super activateIn: swarmContext];

  // Now activate our own schedule.
  [modelSchedule activateIn: self];

  // Finally, return our activity.
  return [self getSwarmActivity];
}

-doStats
{
    //  kludged 
    Quad* quad;
    ResourceQuad* resquad;
    Critter* plant;
    // double fedpercent;
    int now;
    ResList* inflow;
    
    now = getCurrentTime();
    inflow = [landRoot fedOutP];
    
//   printLog("Round %d: fed out ", 
//        getCurrentTime()); 
//    fedpercent = 100.0 * inflow->list[sunwater];
//    printLogD("%7.2f%%\n", fedpercent);

    // biomass
    [inflow drop];
    inflow = [landRoot fedOut];
    [species forEach: M(zeroBiomass)];
    
    quad = [plantRoot pickWin: plantRoot->mcr Root: plantRoot Invisibles: 0];
    while (quad) {
        plant = quad->data;
        [plant->species addBiomass: [plant getVolume]];
        [plant->species tallyAlive: plant];
        quad = [quad pickWin: plantRoot->mcr Root: plantRoot Invisibles: 0];
    }

    quad = [mobileRoot pickWin: mobileRoot->mcr Root: plantRoot Invisibles: 0];
    while (quad) {
        plant = quad->data;
        [plant->species addBiomass: [plant getVolume]];
        [plant->species tallyAlive: plant];
        quad = [quad pickWin: mobileRoot->mcr Root: plantRoot Invisibles: 0];
    }

    resquad = (ResourceQuad*) [landRoot pickWin: landRoot->mcr 
                               Root: landRoot Invisibles: 0];
    while (resquad) {
        [resquad->siteDef setBiomass: [resquad getReservoir]];
        [resquad->siteDef addBiomassFed: [resquad fedOut]];
        [resquad zeroFed];
        resquad = (ResourceQuad*) [resquad pickWin: landRoot->mcr 
                                   Root: landRoot Invisibles: 0];
    }
    
    // now scale 'em
    [species forEach: M(scaleBiomass)];
    [sites forEach: M(scaleBiomass)];
    
    [species forEach: M(ageUpdate)];
    [species forEach: M(sizeUpdate)];
    [species forEach: M(stageUpdate)];
    [species forEach: M(specificUpdate)];
    [species forEach: M(biomassRateUpdate)];
    
    // now do whatever to prepare for next statcycle
    // must do after all stats calculated since some depend on others
    [species forEach: M(allStatsReset)];

    flushBoth();
    
    return self;
}

-(BOOL) extinctCheck
    // anybody gone extinct yet?
{
    id index;
    Species* onespecies;

    index = [species begin: [self getZone]];
    while ( (onespecies = [index next] ) ) {
        if (    ([onespecies getLastAlive] < getCurrentTime()) 
            && (getCurrentTime() > [onespecies getBirthday] ) ) {
            if ([geckoControl getESRun]) {
                [geckoControl stopRunning];
            }
            return 1;
        }
    }
    return 0;
}

-runFitness: (double*) fit
    // Simplest function, sum of % of run length that each species survived.
    // max = #species. (All made it to 100% of run length.)
{
    id index;
    Species* onespecies;

    curRunFitness = 0;
    
    index = [species begin: [self getZone]];
    while ( (onespecies = [index next] ) ) {
        curRunFitness += (double)[onespecies getLastAlive];
    }
    curRunFitness /= (double) settings->stopday;
    
    *fit = curRunFitness;
    return self;
}

@end
