// Code to test dynamic method calls, inspired by the code
// in libobjects mframe.m (But modified to take more advantage of the
// existing runtime info).

#import <objc/Object.h>
#import <objc/objc.h>
#import <objc/objc-api.h>
#import <objc/encoding.h>
#import <stdio.h>
#import <string.h>
#import <stdlib.h>

// These two functions are borrowed from libobjects. The stack args
// one is more or less in the runtime, but the register args one isn't
int
method_types_get_size_of_stack_arguments (const char *type) {
  type = objc_skip_typespec (type);
  return atoi (type);
}

int
method_types_get_size_of_register_arguments(const char *types) {
  const char* type = strrchr(types, '+');
  if (type)
    return atoi(++type) + sizeof(void*);
  else
    return 0;
}                           

// Construct a call on the fly using __builtin_apply.
// Note that we're doing this *without* a varargs frame being passed in,
// That frame is being made here by hand in the variable "argFrame"
// Issues:
//   Why can't I get the type from the selector?
//   Can only return 32 bit values (see mframe for solution).
//   alloca() might not return storage aligned properly for all
//     arguments that are being copied. In particular, on a Sparc if a
//     double needs to be written at the wrong address (say,
//     0xeffff84c), it seems to crash.
//   mframe.m calls objc_skip_type_qualifiers before reading in the
//     argument type - is this necessary? Try out consts..
//   Current code is for demo purposes only, only some datatypes are
//     supported and their values are hardwired in. A useful version of
//     this needs to decode the arguments from elsewhere.

int
constructCall(id object, char * selName) {
  SEL sel;
  const char * selType;
  Method * method;
  IMP imp;
  arglist_t argFrame;
  int numArgs, stackArgSize, regArgSize;
  const char * typePointer;
  void * arg;
  retval_t returnValue;
  int filledArgs;
  
  // get the selector from the string
  sel = sel_get_uid(selName);
  if (sel == 0)
    printf("Selector %s is null\n", selName);

  // Now get the method struct from the selector.
  method = class_get_instance_method(object->class_pointer, sel);
  if (method == 0) {
    printf("Method for selector %s not found.\n", selName);
  }

  // and get the type from the selector (doesn't work?)
  selType = sel_get_type(sel);
// Ginger:  decided to see what this does, then do the other way.
  if (selType)
    printf("sel_get_type: Selector %s has type %s\n", selName, selType);
  else
    printf("sel_get_type: Error getting type for selector %s\n", selName);

  // use the method to get the type
  selType = method->method_types;
  if (selType)
    printf("method->method_types: Selector %s has type %s\n", selName, selType);
  else
    printf("method->method_types: Error getting type for selector %s\n", selName);

  // And read useful data from the method struct (and types)
  numArgs = method_get_number_of_arguments(method);
  stackArgSize = method_types_get_size_of_stack_arguments(selType);
  regArgSize = method_types_get_size_of_register_arguments(selType);
  printf("Selector %s takes %d args, %d bytes on the stack, %d bytes of registers\n", selName, numArgs, stackArgSize, regArgSize);
  
  // Make an argument frame: this stores stack args and register args
  // reg_argsize is how many bytes are needed to store all the registers
  // (do we need to use alloca here, or is malloc ok if we free later?)
  argFrame = (arglist_t) alloca(sizeof(char *) + regArgSize);
  printf("Allocated argFrame of %d bytes at %p\n",
         sizeof(char *) + regArgSize, argFrame);
  
  // Now allocate extra space for the stack args, if need be.
  if (stackArgSize > 0) {
    argFrame->arg_ptr = alloca(stackArgSize);
    printf("Allocated stack argFrame of %d bytes at %p\n",
           stackArgSize, argFrame->arg_ptr);
  } else {
    argFrame->arg_ptr = 0;
    printf("Not allocating any stack argFrame, none needed.\n");
  }

  // Ok, now time to fill in the argFrame. During this, we use some vars:
  //   typePointer keeps track of how much of the type we've consumed
  //     (updated in method_get_next_argument)
  //   arg points to where the argument should be put in the argFrame
  //   filledArgs is counting how many arguments we have actually filled

  typePointer = selType;
  filledArgs = 0;
  printf("Now proceeding to fill in the argument frame.\n");

  // The first two arguments for an objc method are the receiving object 
  // and the selector that was called, so we fill those in by hand.

  // Fill in the receiving object
//  arg = method_get_next_argument(argFrame, &typePointer);
  arg = method_get_first_argument(method, argFrame, &typePointer);
  if (*typePointer == _C_ID)
    printf("  Putting ID %p on the argument list at %p.\n", object, arg);
  else
    printf("  We expected to see an ID, but got typespec %c\n", *typePointer);
  *(id *)arg = object;
  filledArgs++;
  
  // And fill in the selector.
  arg = method_get_next_argument(argFrame, &typePointer);
  if (*typePointer == _C_SEL)
    printf("  Putting selector %p on the argument list at %p.\n", sel, arg);
  else
    printf("  We expected to see a SEL, but got typespec %c\n", *typePointer);
  *(SEL *)arg = sel;
  filledArgs++;

  // OK, now it's time to fill in the rest of the arguments freeform.
  // in this version we only handle a few types, and we hardcode in their
  // values. Real code should do better.
  
  // Ginger - changed method call to nth arg because next arg wasn't filling
  // in correct number -- does nth?
  printf("  ...typepointer = %c at filledArgs = %d\n", *typePointer, filledArgs);
  arg = method_get_nth_argument(method, argFrame, filledArgs++, &typePointer);
  printf("  ...typepointer = %c at filledArgs = %d\n", *typePointer, filledArgs);

//  arg = method_get_next_argument(argFrame, &typePointer);
  while(arg) {
    switch (*typePointer) {                       // should catch all options
      case _C_INT:
        printf("  Putting integer 5 on the argument list at %p.\n", arg);
        *(int *)arg = 5;
        break;
      case _C_DBL:
        printf("  Putting double 2.3 on the argument list at %p.\n", arg);
        *(double *)arg = 2.3;
        break;
      case _C_FLT:
        printf("  Putting float 1.7 on the argument list at %p.\n", arg);
        *(float *)arg = (float) 1.7;
        break;
      case _C_PTR: 
      case _C_CHARPTR:
      case _C_ID:
        printf("  Putting pointer 0x12345678 on the argument list at %p.\n", arg);
        *(void **)arg = (void *)0x12345678;
        break;
      default:
        printf("  Was asked to fill in type %c on the argument list at %p.\n", *typePointer, arg);
    }
    // arg = method_get_next_argument(argFrame, &typePointer);
    arg = method_get_nth_argument(method, argFrame, filledArgs++, &typePointer);
  printf("  ...typepointer = %c at filledArgs = %d\n", *typePointer, filledArgs);
    // filledArgs++;
  }

  if (filledArgs != numArgs) {
    printf("Strange, we expected %d arguments but filled in %d\n",
           numArgs, filledArgs);
  }
       
  // OK, we've got our argument frame built! Now let's make the call.
  // First, lookup the function pointer (IMP)
  imp = objc_msg_lookup(object, sel);
  if (imp == 0)
    printf("Couldn't find imp for selector %s\n", selName);
  else
    printf("Found imp, now making the call.\n");

  // Now make the call! (What happens to register args in the frame?)
  returnValue = __builtin_apply((apply_t)imp, argFrame, stackArgSize);
  __builtin_return(returnValue);
}


// Here's a test class to try various constructed calls to. The idea
// here is to test various alignments of arguments.
@interface TestClass : Object {
}
-(int) mDouble: (const double) d Int: (int) n;
-(int) mInt: (int) d Double: (double) n;
-(int) mPointer: (void *) p Double: (double) d Int: (int) n;
-(int) mPointer: (void *) p Int: (int) n Double: (double) d;
-(int) mDouble: (double) d Int: (int) n Pointer: (void *) p;
-(int) mDouble: (double) d Pointer: (void *) p Int: (int) n;
@end

@implementation TestClass
-(int) mDouble: (const double) d Int: (int) n {
  printf("!! mDouble: %f Int: %d\n", d, n);
  return 1;
}

-(int) mInt: (int) d Double: (double) n {
  printf("!! mInt: %d Double: %f\n", d, n);
  return 2;
}

-(int) mPointer: (void *) p Double: (double) d Int: (int) n {
  printf("!! mPointer: %p Double: %f Int: %d\n", p, d, n);
  return 3;
}

-(int) mPointer: (void *) p Int: (int) n Double: (double) d {
  printf("!! mPointer: %p Int: %d Double: %f\n", p, n, d);
  return 4;
}

-(int) mDouble: (double) d Int: (int) n Pointer: (void *) p {
  printf("!! mDouble: %f Int: %d Pointer: %p\n", d, n, p);
  return 5;
}

-(int) mDouble: (double) d Pointer: (void *) p Int: (int) n {
  printf("!! mDouble: %f Pointer: %p Int: %d\n", d, p, n);
  return 6;
}

@end


// And now, a main function to call all these things, both ways.
int
main() {
  id obj;
  int r1, r2;
  
  obj = [TestClass new];

  r1 = [obj mDouble: 2.3 Int: 5];
  r2 = constructCall(obj, "mDouble:Int:");
  printf("!! Return values: %d %d\n\n", r1, r2);

  r1 = [obj mInt: 5 Double: 2.3];
  r2 = constructCall(obj, "mInt:Double:");
  printf("!! Return values: %d %d\n\n", r1, r2);

  r1 = [obj mPointer: (void *)0x12345678 Double: 2.3 Int: 5];
  r2 = constructCall(obj, "mPointer:Double:Int:");
  printf("!! Return values: %d %d\n\n", r1, r2);
  
  r1 = [obj mPointer: (void *)0x12345678 Int: 5 Double: 2.3];
  r2 = constructCall(obj, "mPointer:Int:Double:");
  printf("!! Return values: %d %d\n\n", r1, r2);
  
  r1 = [obj mDouble: 2.3 Int: 5 Pointer: (void *)0x12345678];
  r2 = constructCall(obj, "mDouble:Int:Pointer:");
  printf("!! Return values: %d %d\n\n", r1, r2);
  
  r1 = [obj mDouble: 2.3 Pointer: (void *)0x12345678 Int: 5];
  r2 = constructCall(obj, "mDouble:Pointer:Int:");
  printf("!! Return values: %d %d\n\n", r1, r2);

  return 0;
}




